﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.ComponentModel;
using System.Linq;

namespace HeuristicLab.Encodings.RealVectorEncoding {
  [Item("CMAES Strategy Parameters", "CMA-ES controls many strategy parameters that guide the search and which are combined in this class.")]
  [StorableClass]
  public sealed class CMAParameters : Item, INotifyPropertyChanged {

    [Storable]
    private DoubleValue axisRatio;
    public DoubleValue AxisRatio {
      get { return axisRatio; }
      set {
        if (axisRatio == value) return;
        axisRatio = value;
        OnPropertyChanged("AxisRatio");
      }
    }

    [Storable]
    private DoubleValue sigma;
    public DoubleValue Sigma {
      get { return sigma; }
      set {
        if (sigma == value) return;
        sigma = value;
        OnPropertyChanged("Sigma");
      }
    }

    [Storable]
    private DoubleArray minSigma;
    public DoubleArray MinSigma {
      get { return minSigma; }
      set {
        if (minSigma == value) return;
        minSigma = value;
        OnPropertyChanged("MinSigma");
      }
    }

    [Storable]
    private DoubleArray maxSigma;
    public DoubleArray MaxSigma {
      get { return maxSigma; }
      set {
        if (maxSigma == value) return;
        maxSigma = value;
        OnPropertyChanged("MaxSigma");
      }
    }

    [Storable]
    private IntValue mu;
    public IntValue Mu {
      get { return mu; }
      set {
        if (mu == value) return;
        mu = value;
        OnPropertyChanged("Mu");
      }
    }

    [Storable]
    private DoubleArray weights;
    public DoubleArray Weights {
      get { return weights; }
      set {
        if (weights == value) return;
        weights = value;
        OnPropertyChanged("Weights");
      }
    }

    [Storable]
    private DoubleValue muEff;
    public DoubleValue MuEff {
      get { return muEff; }
      set {
        if (muEff == value) return;
        muEff = value;
        OnPropertyChanged("MuEff");
      }
    }

    [Storable]
    private DoubleValue cc;
    public DoubleValue CC {
      get { return cc; }
      set {
        if (cc == value) return;
        cc = value;
        OnPropertyChanged("CC");
      }
    }

    [Storable]
    private DoubleValue cs;
    public DoubleValue CS {
      get { return cs; }
      set {
        if (cs == value) return;
        cs = value;
        OnPropertyChanged("CS");
      }
    }

    [Storable]
    private DoubleValue c1;
    public DoubleValue C1 {
      get { return c1; }
      set {
        if (c1 == value) return;
        c1 = value;
        OnPropertyChanged("C1");
      }
    }

    [Storable]
    private DoubleValue cmu;
    public DoubleValue CMU {
      get { return cmu; }
      set {
        if (cmu == value) return;
        cmu = value;
        OnPropertyChanged("CMU");
      }
    }

    [Storable]
    private DoubleValue damps;
    public DoubleValue Damps {
      get { return damps; }
      set {
        if (damps == value) return;
        damps = value;
        OnPropertyChanged("Damps");
      }
    }

    [Storable]
    private DoubleValue muCov;
    public DoubleValue MuCov {
      get { return muCov; }
      set {
        if (muCov == value) return;
        muCov = value;
        OnPropertyChanged("MuCov");
      }
    }

    [Storable]
    private DoubleValue cCov;
    public DoubleValue CCov {
      get { return cCov; }
      set {
        if (cCov == value) return;
        cCov = value;
        OnPropertyChanged("CCov");
      }
    }

    [Storable]
    private DoubleValue cCovSep;
    public DoubleValue CCovSep {
      get { return cCovSep; }
      set {
        if (cCovSep == value) return;
        cCovSep = value;
        OnPropertyChanged("CCovSep");
      }
    }

    [Storable]
    private DoubleArray pc;
    public DoubleArray PC {
      get { return pc; }
      set {
        if (pc == value) return;
        pc = value;
        OnPropertyChanged("PC");
      }
    }

    [Storable]
    private DoubleArray ps;
    public DoubleArray PS {
      get { return ps; }
      set {
        if (ps == value) return;
        ps = value;
        OnPropertyChanged("PS");
      }
    }

    [Storable]
    private DoubleMatrix b;
    public DoubleMatrix B {
      get { return b; }
      set {
        if (b == value) return;
        b = value;
        OnPropertyChanged("B");
      }
    }

    [Storable]
    private DoubleArray d;
    public DoubleArray D {
      get { return d; }
      set {
        if (d == value) return;
        d = value;
        OnPropertyChanged("D");
      }
    }

    [Storable]
    private DoubleMatrix c;
    public DoubleMatrix C {
      get { return c; }
      set {
        if (c == value) return;
        c = value;
        OnPropertyChanged("C");
      }
    }

    [Storable]
    private DoubleArray bDz;
    public DoubleArray BDz {
      get { return bDz; }
      set {
        if (bDz == value) return;
        bDz = value;
        OnPropertyChanged("BDz");
      }
    }

    [Storable]
    private DoubleValue chiN;
    public DoubleValue ChiN {
      get { return chiN; }
      set {
        if (chiN == value) return;
        chiN = value;
        OnPropertyChanged("ChiN");
      }
    }

    [Storable]
    private IntValue initialIterations;
    public IntValue InitialIterations {
      get { return initialIterations; }
      set {
        if (initialIterations == value) return;
        initialIterations = value;
        OnPropertyChanged("InitialIterations");
      }
    }

    [StorableConstructor]
    private CMAParameters(bool deserializing) : base(deserializing) { }
    private CMAParameters(CMAParameters original, Cloner cloner)
      : base(original, cloner) {
      this.axisRatio = cloner.Clone(original.axisRatio);
      this.b = cloner.Clone(original.b);
      this.c = cloner.Clone(original.c);
      this.c1 = cloner.Clone(original.c1);
      this.cc = cloner.Clone(original.cc);
      this.cCov = cloner.Clone(original.cCov);
      this.cCovSep = cloner.Clone(original.cCovSep);
      this.chiN = cloner.Clone(original.chiN);
      this.cmu = cloner.Clone(original.cmu);
      this.cs = cloner.Clone(original.cs);
      this.d = cloner.Clone(original.d);
      this.initialIterations = cloner.Clone(original.initialIterations);
      this.bDz = cloner.Clone(original.bDz);
      this.maxSigma = cloner.Clone(original.maxSigma);
      this.minSigma = cloner.Clone(original.minSigma);
      this.mu = cloner.Clone(original.mu);
      this.muCov = cloner.Clone(original.muCov);
      this.muEff = cloner.Clone(original.muEff);
      this.pc = cloner.Clone(original.pc);
      this.ps = cloner.Clone(original.ps);
      this.sigma = cloner.Clone(original.sigma);
      this.weights = cloner.Clone(original.weights);
    }
    public CMAParameters() { }
    public CMAParameters(int dimension, int lambda) {
      SetDefaults(dimension, lambda);
    }

    private void SetDefaults(int N, int lambda) {
      ChiN = new DoubleValue(Math.Sqrt(N) * (1.0 - 1.0 / (4.0 * N) + 1.0 / (21.0 * N * N)));
      Mu = new IntValue((int)Math.Floor(lambda / 2.0));

      var w = new double[Mu.Value];
      for (int i = 0; i < Mu.Value; i++)
        w[i] = (Math.Log(Mu.Value + 1) - Math.Log(i + 1));

      var sum = w.Sum();
      for (int i = 0; i < w.Length; i++) w[i] /= sum;

      Weights = new DoubleArray(w);
      MuEff = new DoubleValue(1.0 / Weights.Sum(x => x * x));
      CS = new DoubleValue((MuEff.Value + 2) / (N + MuEff.Value + 3));
      Damps = new DoubleValue(2 * Math.Max(0, Math.Sqrt((MuEff.Value - 1) / (N + 1)) - 1)
                                * Math.Max(0.3, 1 - N / (1e-6 + /*TODO earliest stop (MaxGen)*/ 1000))
                              + CS.Value + 1);
      CC = new DoubleValue(4.0 / (N + 4));
      MuCov = new DoubleValue(MuEff.Value);
      CCov = new DoubleValue(2.0 / ((N + 1.41) * (N + 1.41) * MuCov.Value)
                             + (1 - (1.0 / MuCov.Value)) * Math.Min(1, (2 * MuEff.Value - 1) / (MuEff.Value + (N + 2) * (N + 2))));
      CCovSep = new DoubleValue(Math.Min(1, CCov.Value * (N + 1.5) / 3));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CMAParameters(this, cloner);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string name) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(name));
    }
  }
}
