﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.Encodings.RealVectorEncoding {
  [Item("CMARecombinator", "Calculates the weighted mean of a number of offspring.")]
  [StorableClass]
  public class CMARecombinator : SingleSuccessorOperator, IRealVectorOperator, ICMAESRecombinator {

    public Type CMAType {
      get { return typeof(CMAParameters); }
    }

    #region Parameter Properties
    public IScopeTreeLookupParameter<RealVector> OffspringParameter {
      get { return (IScopeTreeLookupParameter<RealVector>)Parameters["Offspring"]; }
    }

    public ILookupParameter<RealVector> MeanParameter {
      get { return (ILookupParameter<RealVector>)Parameters["Mean"]; }
    }

    public ILookupParameter<CMAParameters> StrategyParametersParameter {
      get { return (ILookupParameter<CMAParameters>)Parameters["StrategyParameters"]; }
    }
    #endregion

    [StorableConstructor]
    protected CMARecombinator(bool deserializing) : base(deserializing) { }
    protected CMARecombinator(CMARecombinator original, Cloner cloner) : base(original, cloner) { }
    public CMARecombinator()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<RealVector>("Offspring", "The offspring that should be recombined."));
      Parameters.Add(new LookupParameter<RealVector>("Mean", "The new mean solution."));
      Parameters.Add(new LookupParameter<CMAParameters>("StrategyParameters", "The CMA-ES strategy parameters used for mutation."));
      OffspringParameter.ActualName = "RealVector";
      MeanParameter.ActualName = "RealVector";
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CMARecombinator(this, cloner);
    }

    public override IOperation Apply() {
      var sp = StrategyParametersParameter.ActualValue;
      var offspring = OffspringParameter.ActualValue;
      var mean = new RealVector(offspring[0].Length);
      for (int i = 0; i < mean.Length; i++) {
        for (int j = 0; j < offspring.Length; j++)
          mean[i] += sp.Weights[j] * offspring[j][i];
      }

      MeanParameter.ActualValue = mean;
      return base.Apply();
    }
  }
}